@@ -1,9 +1,9 @@
 /*
- * $Id: ngrep.c,v 1.32 2003/08/10 04:14:52 jpr5 Exp $
+ * $Id: ngrep.c,v 1.46 2004/03/29 02:15:59 jpr5 Exp $
  *
- * Copyright (c) 2001  Jordan Ritter <jpr5@darkridge.com>
+ * Copyright (c) 2004  Jordan Ritter <jpr5@darkridge.com>
  *
- * Please refer to the COPYRIGHT file for more information.
+ * Please refer to the LICENSE file for more information.
  *
  */
 
@@ -60,9 +60,15 @@
 #include <stdlib.h>
 #include <string.h>
 #include <signal.h>
+#include <errno.h>
 #include <sys/ioctl.h>
 
-#ifdef USE_PCRE
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#if USE_PCRE
 #include "pcre-3.4/pcre.h"
 #else
 #include "regex-0.12/regex.h"
@@ -71,7 +77,7 @@
 #include "ngrep.h"
 
 
-static char rcsver[] = "$Revision: 1.32 $";
+static char rcsver[] = "$Revision: 1.46 $";
 
 int snaplen = 65535, limitlen = 65535, promisc = 1, to = 1000;
 int show_empty = 0, show_hex = 0, quiet = 0;
@@ -79,26 +85,30 @@ int match_after = 0, keep_matching = 0;
 int invert_match = 0, bin_match = 0;
 int matches = 0, max_matches = 0;
 int live_read = 1, want_delay = 0;
+int no_dropprivs = 0;
+
+char nonprint_char = '.';
 
 char pc_err[PCAP_ERRBUF_SIZE];
-#ifdef USE_PCRE
+#if USE_PCRE
 int err_offset;
 char *re_err = NULL;
 #else
 const char *re_err = NULL;
 #endif
 
-int re_match_word = 0, re_ignore_case = 0;
+int re_match_word = 0, re_ignore_case = 0, re_multiline_match = 1;
 
-#ifdef USE_PCRE
+#if USE_PCRE
 pcre *pattern = NULL;
 pcre_extra *pattern_extra = NULL;
 #else
 struct re_pattern_buffer pattern;
 #endif
 
-char *match_data = NULL, *bin_data = NULL, *filter = NULL;
+char *match_data = NULL, *bin_data = NULL, *filter = NULL, *filter_file = NULL;
 int (*match_func)() = &blank_match_func;
+void (*dump_func)(char *, int) = &dump_formatted;
 int match_len = 0;
 
 struct bpf_program pcapfilter;
@@ -113,20 +123,39 @@ pcap_dumper_t *pd_dump = NULL;
 struct timeval prev_ts = {0, 0}, prev_delay_ts = {0,0};
 void (*print_time)() = NULL, (*dump_delay)() = dump_delay_proc_init;
 
-unsigned ws_row, ws_col;
+unsigned ws_row, ws_col, ws_col_forced = 0;
 
 
 int main(int argc, char **argv) {
     int c;
 
-    signal(SIGINT,  clean_exit);
-    signal(SIGQUIT, clean_exit);
-    signal(SIGABRT, clean_exit);
-    signal(SIGPIPE, clean_exit);
+    signal(SIGINT,   clean_exit);
+    signal(SIGQUIT,  clean_exit);
+    signal(SIGABRT,  clean_exit);
+    signal(SIGPIPE,  clean_exit);
     signal(SIGWINCH, update_windowsize);
 
-    while ((c = getopt(argc, argv, "hXViwqpevxlDtTs:n:d:A:I:O:S:")) != EOF) {
+    while ((c = getopt(argc, argv, "hXViwqpevxlDtTRMs:n:c:d:A:I:O:S:P:F:W:")) != EOF) {
         switch (c) {
+            case 'W': {
+                if (!strcasecmp(optarg, "normal"))
+                    dump_func = &dump_formatted;
+                else if (!strcasecmp(optarg, "byline"))
+                    dump_func = &dump_byline;
+                else if (!strcasecmp(optarg, "none"))
+                    dump_func = &dump_unwrapped;
+                else {
+                    printf("fatal: unknown wrap method '%s'\n", optarg);
+                    usage(-1);
+                }
+            } break;
+
+            case 'F':
+                filter_file = optarg;
+                break;
+            case 'P':
+                nonprint_char = *optarg;
+                break;
             case 'S':
                 limitlen = atoi(optarg);
                 break;
@@ -142,12 +171,21 @@ int main(int argc, char **argv) {
             case 'd':
                 dev = optarg;
                 break;
+            case 'c':
+                ws_col_forced = atoi(optarg);
+                break;
             case 'n':
                 max_matches = atoi(optarg);
                 break;
             case 's':
                 snaplen = atoi(optarg);
                 break;
+            case 'M':
+                re_multiline_match = 0;
+                break;
+            case 'R':
+                no_dropprivs = 1;
+                break;
             case 'T':
                 print_time = &print_time_diff_init;
                 break;
@@ -193,11 +231,16 @@ int main(int argc, char **argv) {
         }
     }
 
+    if (show_hex && dump_func != &dump_formatted) {
+        printf("fatal: -x (hex dump) is incompatible with -W (alternate format)\n");
+        usage(-1);
+    }
 
     if (argv[optind])
         match_data = argv[optind++];
 
     if (read_file) {
+
         if (!(pd = pcap_open_offline(read_file, pc_err))) {
             perror(pc_err);
             clean_exit(-1);
@@ -208,8 +251,8 @@ int main(int argc, char **argv) {
         live_read = 0;
         printf("input: %s\n", read_file);
 
-
     } else {
+
         if (!dev)
             if (!(dev = pcap_lookupdev(pc_err))) {
                 perror(pc_err);
@@ -239,23 +282,42 @@ int main(int argc, char **argv) {
         }
     }
 
+    if (filter_file) {
+        char buf[1024] = {0};
+        FILE *f = fopen(filter_file, "r");
 
-    if (argv[optind]) {
-        filter = get_filter(&argv[optind]);
+        if (!f || !fgets(buf, sizeof(buf)-1, f)) {
+            fprintf(stderr, "fatal: unable to get filter from %s: %s\n", filter_file, strerror(errno));
+            usage(-1);
+        }
+
+        fclose(f);
+
+        filter = get_filter_from_string(buf);
+
+        if (pcap_compile(pd, &pcapfilter, filter, 0, mask.s_addr)) {
+            pcap_perror(pd, "pcap compile");
+            clean_exit(-1);
+        }
+
+    } else if (argv[optind]) {
+        filter = get_filter_from_argv(&argv[optind]);
 
         if (pcap_compile(pd, &pcapfilter, filter, 0, mask.s_addr)) {
             free(filter);
-            filter = get_filter(&argv[optind-1]);
+            filter = get_filter_from_argv(&argv[optind-1]);
 
-#ifdef NEED_RESTART
-            PCAP_RESTART();
+#if USE_PCAP_RESTART
+            PCAP_RESTART_FUNC();
 #endif
             if (pcap_compile(pd, &pcapfilter, filter, 0, mask.s_addr)) {
                 pcap_perror(pd, "pcap compile");
                 clean_exit(-1);
             } else match_data = NULL;
         }
+    }
 
+    if (filter) {
         if (!quiet) printf("filter: %s\n", filter);
 
         if (pcap_setfilter(pd, &pcapfilter)) {
@@ -300,15 +362,20 @@ int main(int argc, char **argv) {
 
         } else {
 
-#ifdef USE_PCRE
+#if USE_PCRE
             int pcre_options = PCRE_UNGREEDY;
 
             if (re_ignore_case)
                 pcre_options |= PCRE_CASELESS;
 
-            re_err = malloc(512);
+            if (re_multiline_match)
+                pcre_options |= PCRE_DOTALL;
 #else
-            re_syntax_options = RE_SYNTAX_EGREP;
+            re_syntax_options = RE_CHAR_CLASSES | RE_NO_BK_PARENS | RE_NO_BK_VBAR |
+                                RE_CONTEXT_INDEP_ANCHORS | RE_CONTEXT_INDEP_OPS;
+
+            if (re_multiline_match)
+                re_syntax_options |= RE_DOT_NEWLINE;
 
             if (re_ignore_case) {
                 char *s;
@@ -316,6 +383,7 @@ int main(int argc, char **argv) {
 
                 pattern.translate = (char*)malloc(256);
                 s = pattern.translate;
+
                 for (i = 0; i < 256; i++)
                     s[i] = i;
                 for (i = 'A'; i <= 'Z'; i++)
@@ -324,6 +392,7 @@ int main(int argc, char **argv) {
                 s = match_data;
                 while (*s)
                     *s++ = tolower(*s);
+
             } else pattern.translate = NULL;
 #endif
 
@@ -333,17 +402,15 @@ int main(int argc, char **argv) {
                 match_data = word_regex;
             }
 
-#ifdef USE_PCRE
+#if USE_PCRE
             pattern = pcre_compile(match_data, pcre_options, (const char **)&re_err, &err_offset, 0);
+
             if (!pattern) {
                 fprintf(stderr, "compile failed: %s\n", re_err);
                 clean_exit(-1);
             }
 
             pattern_extra = pcre_study(pattern, 0, (const char **)&re_err);
-
-            free(re_err);
-            re_err = NULL;
 #else
             re_err = re_compile_pattern(match_data, strlen(match_data), &pattern);
             if (re_err) {
@@ -392,24 +459,26 @@ int main(int argc, char **argv) {
             link_offset = PPPHDR_SIZE;
             break;
 
-        case DLT_RAW:
-            link_offset = RAWHDR_SIZE;
-            break;
-
-#ifdef HAVE_LOOP
+#if HAVE_DLT_LOOP
         case DLT_LOOP:
 #endif
         case DLT_NULL:
             link_offset = LOOPHDR_SIZE;
             break;
 
-#ifdef HAVE_SLL
+#if HAVE_DLT_RAW
+        case DLT_RAW:
+            link_offset = RAWHDR_SIZE;
+            break;
+#endif
+
+#if HAVE_DLT_LINUX_SLL
         case DLT_LINUX_SLL:
             link_offset = ISDNHDR_SIZE;
             break;
 #endif
 
-#ifdef HAVE_802_11
+#if HAVE_DLT_IEEE802_11
         case DLT_IEEE802_11:
             link_offset = IEEE80211HDR_SIZE;
             break;
@@ -484,7 +553,7 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
                 if (tcphdr_offset || !frag_offset) {
                     printf("%s:%d -", inet_ntoa(ip_packet->ip_src), ntohs(tcp->th_sport));
                     printf("> %s:%d", inet_ntoa(ip_packet->ip_dst), ntohs(tcp->th_dport));
-                    printf(" [%s%s%s%s%s%s]",
+                    printf(" [%s%s%s%s%s%s%s%s]",
                            (tcp->th_flags & TH_ACK)?"A":"",
                            (tcp->th_flags & TH_SYN)?"S":"",
                            (tcp->th_flags & TH_RST)?"R":"",
@@ -503,13 +572,13 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
                            frag_offset, len);
                 else printf("\n");
 
-                if (pd_dump) {
+                if (pd_dump)
                     pcap_dump((u_char*)pd_dump, h, p);
-                    if (!quiet) dump(data, len);
-                } else dump(data, len);
+
+                if (quiet < 2)
+                    dump_func(data, len);
             }
-        }
-            break;
+        } break;
 
         case IPPROTO_UDP: {
             struct udphdr* udp = (struct udphdr *)(((char *)ip_packet) + ip_hl);
@@ -540,7 +609,7 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
                     print_time(h);
 
                 if (udphdr_offset || !frag_offset) {
-#ifdef HAVE_DUMB_UDPHDR
+#if HAVE_DUMB_UDPHDR
                     printf("%s:%d -", inet_ntoa(ip_packet->ip_src), ntohs(udp->source));
                     printf("> %s:%d", inet_ntoa(ip_packet->ip_dst), ntohs(udp->dest));
 #else
@@ -557,13 +626,13 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
                            frag_offset, len);
                 else printf("\n");
 
-                if (pd_dump) {
+                if (pd_dump)
                     pcap_dump((u_char*)pd_dump, h, p);
-                    if (!quiet) dump(data, len);
-                } else dump(data, len);
+
+                if (quiet < 2)
+                    dump_func(data, len);
             }
-        }
-            break;
+        } break;
 
         case IPPROTO_ICMP: {
             struct icmp* ic = (struct icmp *)(((char *)ip_packet) + ip_hl);
@@ -604,13 +673,13 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
                            frag_offset, len);
                 else printf("\n");
 
-                if (pd_dump) {
+                if (pd_dump)
                     pcap_dump((u_char*)pd_dump, h, p);
-                    if (!quiet) dump(data, len);
-                } else dump(data, len);
+
+                if (quiet < 2)
+                    dump_func(data, len);
             }
-        }
-            break;
+        } break;
 
     }
 
@@ -620,7 +689,7 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 
 
 int re_match_func(char *data, int len) {
-#ifdef USE_PCRE
+#if USE_PCRE
     switch(pcre_exec(pattern, 0, data, len, 0, 0, 0, 0)) {
         case PCRE_ERROR_NULL:
         case PCRE_ERROR_BADOPTION:
@@ -684,7 +753,35 @@ int blank_match_func(char *data, int len) {
 }
 
 
-void dump(char *data, int len) {
+void dump_byline(char *data, int len) {
+    if (len > 0) {
+        const char *s = data;
+        unsigned width;
+
+        while (s < data + len) {
+            printf("%c", (*s == '\n' || isprint(*s))? *s : nonprint_char);
+            s++;
+        }
+
+        printf("\n");
+    }
+}
+
+void dump_unwrapped(char *data, int len) {
+    if (len > 0) {
+        const char *s = data;
+        unsigned width;
+
+        while (s < data + len) {
+            printf("%c", isprint(*s) ? *s : nonprint_char);
+            s++;
+        }
+
+        printf("\n");
+    }
+}
+
+void dump_formatted(char *data, int len) {
     if (len > 0) {
         unsigned width = show_hex?16:(ws_col-5);
         char *str = data;
@@ -705,7 +802,7 @@ void dump(char *data, int len) {
 
             for (j = 0; j < width; j++)
                 if (i+j < len)
-                    printf("%c", isprint(str[j])?str[j]:'.');
+                    printf("%c", isprint(str[j]) ? str[j] : nonprint_char);
                 else printf(" ");
 
             str += width;
@@ -716,8 +813,33 @@ void dump(char *data, int len) {
     }
 }
 
+char *get_filter_from_string(char *str) {
+    char *mine;
+    int len;
 
-char *get_filter(char **argv) {
+    if (!str || !*str)
+        return NULL;
+
+    len = strlen(str);
+
+    {
+        char *s;
+        for (s = str; *s; s++)
+            if (*s == '\r' || *s == '\n')
+                *s = ' ';
+    }
+
+    if (!(mine = (char*)malloc(len + sizeof(IP_ONLY))))
+        return NULL;
+
+    memset(mine, 0, len + sizeof(IP_ONLY));
+
+    sprintf(mine, IP_ONLY, str);
+
+    return mine;
+}
+
+char *get_filter_from_argv(char **argv) {
     char **arg = argv, *theirs, *mine;
     char *from, *to;
     int len = 0;
@@ -832,32 +954,97 @@ void dump_delay_proc(struct pcap_pkthdr *h) {
 
 
 void update_windowsize(int e) {
-    const struct winsize ws;
+    if (e == 0 && ws_col_forced)
 
-    if (!ioctl(0, TIOCGWINSZ, &ws)) {
-        ws_row = ws.ws_row;
-        ws_col = ws.ws_col;
-    } else {
-        ws_row = 24;
-        ws_col = 80;
+        ws_col = ws_col_forced;
+
+    else if (!ws_col_forced) {
+        const struct winsize ws;
+
+        if (!ioctl(0, TIOCGWINSZ, &ws)) {
+            ws_row = ws.ws_row;
+            ws_col = ws.ws_col;
+        } else {
+            ws_row = 24;
+            ws_col = 80;
+        }
     }
 }
 
 
 void drop_privs(void) {
-    struct passwd *pw = getpwnam(SAFE_USER);
+#if !USE_DROPPRIVS
+    return;
+#endif
 
-    if (setregid(pw->pw_gid, pw->pw_gid) == -1 ||
-        setreuid(pw->pw_uid, pw->pw_uid) == -1) {
+    if (no_dropprivs)
+        return;
+
+#if DROPPRIVS_ONLY_ROOT
+    /*
+     * We have a choice as to whether we drop privs only if the user
+     * is absolutely root, or if any root privilege in part or whole
+     * is present.  It would be implausible at best to devise logic
+     * that would be valid for every individual semantic of all
+     * supported OSes, i.e. certain syscalls on certain platforms
+     * might or might not rely on real vs. effective privileges to
+     * allow the call to complete.  So, we take the safer route here
+     * which is to drop privs if any root privilege exists.
+     */
+
+    if (!(getuid() == 0 || geteuid() == 0 ||
+          getgid() == 0 || getegid() == 0))
+        return;
+#endif
+
+ {
+    struct passwd *pw = getpwnam(DROPPRIVS_USER);
+    gid_t newgid = pw->pw_uid, oldgid = getegid();
+    uid_t newuid = pw->pw_gid, olduid = geteuid();
+
+    if (!olduid)
+        setgroups(1, &newgid);
+
+    if (newgid != oldgid) {
+#if !defined(LINUX)
+        setegid(newgid);
+        if (setgid(newgid) == -1)
+#else
+        if (setregid(newgid, newgid) == -1)
+#endif
+        {
+            perror("attempt to drop privileges failed");
+            clean_exit(-1);
+        }
+    }
+
+    if (newuid != olduid) {
+#if !defined(LINUX)
+#if !defined(MACOSX)
+        seteuid(newuid);
+#endif
+        if (setuid(newuid) == -1)
+#else
+        if (setreuid(newuid, newuid) == -1)
+#endif
+        {
+            perror("attempt to drop privileges failed");
+            clean_exit(-1);
+        }
+    }
+
+    if ((newgid != oldgid && (setegid(oldgid) != -1 || getegid() != newgid)) ||
+        (newuid != olduid && (seteuid(olduid) != -1 || geteuid() != newuid))) {
         perror("attempt to drop privileges failed");
         clean_exit(-1);
     }
+ }
 }
 
 void usage(int e) {
-    printf("usage: ngrep <-hXViwqpevxlDtT> <-IO pcap_dump> <-n num> <-d dev> <-A num>\n"
-           "                               <-s snaplen> <-S limitlen> <match expression>\n"
-           "                               <bpf filter>\n");
+    printf("usage: ngrep <-hXViwqpevxlDtTRM> <-IO pcap_dump> <-n num> <-d dev> <-A num>\n"
+           "                <-s snaplen> <-S limitlen> <-W normal|byline|none> <-c cols>\n"
+           "                <-P char> <-F file> <match expression> <bpf filter>\n");
 
     exit(e);
 }
@@ -873,8 +1060,7 @@ void clean_exit(int sig) {
     struct pcap_stat s;
     if (!quiet && sig >= 0) printf("exit\n");
 
-#ifdef USE_PCRE
-    if (re_err) free(re_err);
+#if USE_PCRE
     if (pattern) pcre_free(pattern);
     if (pattern_extra) pcre_free(pattern_extra);
 #else
